==> src-bfagent-7.0.2.0-2-0046.tar.gz: bfagent-7.0.2.0-2-0065/src/pty.c <==
/*
Copyright (C)2003-2006, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"
#ifdef HAVE_PTY_H
#include <pty.h>
#endif
#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif
#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif
#ifdef HAVE_SYS_PTY_H
#include <sys/pty.h>
#endif
#ifdef HAVE_TERMIO_H
#include <termio.h>
#endif
#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif
#ifdef HAVE_UTIL_H
#include <util.h>
#endif
#ifdef HAVE_UTMP_H
#include <utmp.h>
#endif



#ifdef DEBUG_PTY
bf_command_state_t command_state;
static int use_pipe(void) { return 0; }
static int create_pipe(void) { return -1; }
#else	/* !DEBUG_PTY */
extern bf_command_state_t command_state;

static int use_pipe(void) {
	return BFCONF_EXISTS(no_pty);
}

static int create_pipe(void) {
	bf_command_state_t *state = &command_state;
	int fds[2];

	strcpy(state->ptyname, bfCat(M_PTY_PIPE));
	if (pipe(fds) != 0)
		return -1;
	state->mfd = fds[0];
	state->sfd = fds[1];
	return 0;
}
#endif	/* DEBUG_PTY */
 


#ifdef HAVE_OPENPTY
int bf_openpty(void) {
	bf_command_state_t *state = &command_state;
	int ret;

	if (use_pipe())
		return create_pipe();

	ret = openpty(&state->mfd, &state->sfd, state->ptyname, NULL, NULL);
	convert_native_to_utf8(state->ptyname, -1, state->buf, BFBUFSIZ);
	strcpy(state->ptyname, state->buf);
	return ret;
}


#else /* !HAVE_OPENPTY */

#ifdef ZOS
/*
 * Unfortunately, finding a pseudo-terminal on z/OS is painful.
 * It uses the old BSD-style pseudo-terminals, where you use
 * brute force to find an available master.
 */
int bf_openpty(void) {
	int *mfd = &command_state.mfd;
	int *sfd = &command_state.sfd;
	char *ptyname = command_state.ptyname;
	int i;

	if (use_pipe())
		return create_pipe();

	for (i=0; i<=9999; i++) {
		/* Look for a master that is available */
		sprintf(ptyname, "/dev/ptyp%04d", i);
		if ((*mfd = open(ptyname, O_RDWR)) < 0) {
			if (errno == ENOENT)
				return -1;
			continue;
		}

		/* See if the matching slave is available */
		ptyname[5] = 't';
		chown(ptyname, getuid(), -1);
		/* It should be appropriate to change this to mode 620, but
		 * the OS is not restoring it to 666 even though it is being
		 * smart enough to restore it to system ownership.
		 */ 
		/* chmod(ptyname, S_IRUSR | S_IWUSR | S_IWGRP); */
		if ((*sfd = open(ptyname, O_RDWR)) < 0) {
			/* Oops, got the master but not the slave */
			close(*mfd);
			continue;
		}

#ifdef HAVE_STROPTS_H
		if (ioctl(*sfd, I_LIST, NULL) > 0) {	/* streams support ? */
			(void)ioctl(*sfd, I_PUSH, "ptem");
			(void)ioctl(*sfd, I_PUSH, "ldterm");
		}
#endif /* HAVE_STROPTS_H */
		return 0;
	}

	/* We exhausted the entire pseudo-terminal namespace?! */
	errno = ENOENT;
	return -1;
}

#else /* !ZOS */

int bf_openpty(void) {
	int *mfd = &command_state.mfd;
	int *sfd = &command_state.sfd;
	char *ptyname = command_state.ptyname;
	int err, cloned = 0;
	char *sname;
#ifdef TIOCTRAP
	int one = 1;
#endif

	if (use_pipe())
		return create_pipe();

	if ((*mfd = open("/dev/ptmx", O_RDWR)) < 0 &&
			(*mfd = open("/dev/ptc", O_RDWR)) < 0)
	{
		if ((*mfd = open("/dev/ptym/clone", O_RDWR)) < 0)
			return -1;

		cloned = 1;
#ifdef TIOCTRAP
		ioctl(*mfd, TIOCTRAP, &one);
#endif
	}

	if ((sname = ptsname(*mfd)) != NULL) {
		strcpy(ptyname, sname);
		if (cloned || (!grantpt(*mfd) && !unlockpt(*mfd))) {
			if ((*sfd = open(ptyname, O_RDWR)) >= 0) {
#ifdef HAVE_STROPTS_H
				/* streams support? */
				if (!cloned && ioctl(*sfd, I_LIST, NULL) > 0) {
					if (ioctl(*sfd, I_FIND, "ldterm") != 1 &&
						ioctl(*sfd, I_FIND, "ldtty") != 1)
					{
						ioctl(*sfd, I_PUSH, "ptem");
						ioctl(*sfd, I_PUSH, "ldterm");
					}
				}
#endif /* HAVE_STROPTS_H */
				return 0;
			}
		}
	}

	err = errno;
	close(*mfd);
	*mfd = -1;
	errno = err;
	return -1;
}

#endif /* ZOS */
#endif /* HAVE_OPENPTY */


#ifdef DEBUG_PTY

int main(int argc, char *argv[]) {
	bf_command_state_t *state = &command_state;
	char **execArgv;
	int i, ret;

	if (argc < 3) {
		printf("Usage: %s command argv0 [argv1 [...]]\n", argv[0]);
		exit(1);
	}

	execArgv = calloc(argc, sizeof(char *));
	for (i=2; i<argc; i++)
		execArgv[i-2] = argv[i];
	execArgv[argc-2] = NULL;

	memset(state, 0, sizeof(*state));
	state->mfd = state->sfd = -1;
	strcpy(state->ptyname, "(unchanged)");

	ret = bf_openpty();
	if (ret != 0) {
		printf("bf_openpty failed: mfd=%d sfd=%d ptyname=%s ret=%d errno=%d\n",
			state->mfd, state->sfd, state->ptyname, ret, errno);
		exit(2);
	}

	execv(argv[1], execArgv);
	printf("exec failed: %s: errno=%d\n", argv[1], errno);
	exit(3);
}

#endif /* DEBUG_PTY */

==> src-bfagent-7.1.1.3-0-0034.tar.gz: bfagent-7.1.1.3-0-0034/src/pty.c <==
/*
Copyright (C)2003-2008, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"
#ifdef HAVE_PTY_H
#include <pty.h>
#endif
#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif
#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif
#ifdef HAVE_SYS_PTY_H
#include <sys/pty.h>
#endif
#ifdef HAVE_TERMIO_H
#include <termio.h>
#endif
#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif
#ifdef HAVE_UTIL_H
#include <util.h>
#endif
#ifdef HAVE_UTMP_H
#include <utmp.h>
#endif


#ifndef O_NOCTTY
#  define O_NOCTTY (0)
#endif /* O_NOCTTY */


void bf_notty(void) {
#ifdef TIOCNOTTY
	int fd;
#  ifdef L_ctermid
	char devtty[L_ctermid];
	ctermid(devtty);
#   else /* !L_ctermid */
	char devtty[16];
	strcpy(devtty, "/dev/tty");
#  endif /* L_ctermid/! */
	fd = open(devtty, O_RDWR | O_NOCTTY);
	if (fd < 0) { return; }
	ioctl(fd, TIOCNOTTY, NULL);
#endif
}



#ifdef DEBUG_PTY
bf_command_state_t command_state;
static int use_pipe(void) { return 0; }
static int create_pipe(void) { return -1; }
#else	/* !DEBUG_PTY */
extern bf_command_state_t command_state;

static int use_pipe(void) {
	return BFCONF_EXISTS(no_pty);
}

static int create_pipe(void) {
	bf_command_state_t *state = &command_state;
	int fds[2];

	strcpy(state->ptyname, "|");
	if (pipe(fds) != 0)
		return -1;
	state->mfd = fds[0];
	state->sfd = fds[1];
	return 0;
}
#endif	/* DEBUG_PTY */
 
/*
 * The behavior of pseudo-terminals is rather different between
 * SVR4- and BSD-derived operating systems.  In particular:
 *
 * On SVR4:
 * - The only thing gauranteed to obtain a new controlling terminal
 * is to open one without the O_NOTTY flag specified while the process
 * does not already have one.
 * - The process can not open a second terminal and arbitrarily switch
 * between them.
 * - The only gauranteed way to get rid of your controlling terminal is
 * setsid(), which will fail if the process is already a process group
 * leader, and the only reliable way to stop being the group leader is
 * to fork().
 *
 * BSD: Pseudo-terminals are never promoted to being the controlling
 * terminal automatically.  The program must use the TIOCSCTTY ioctl on
 * it to bind it.  The TIOCNOTTY ioctl, which reverses this, may or may
 * not be available.
 *
 * The easiest way to deal with all of this is to immediately fork()
 * and setsid() before opening the pty, then calling the TIOCSCTTY ioctl
 * if that symbol is defined.  This function does the fork piece of this
 * process.
 */
static void detach_from_tty(void) {
	int status;
	pid_t pid;

	switch (pid = fork()) {
		case -1:
			/* Fork failed -- make a last, desperate attempt to detach,
			 * but it will almost certainly fail, too...
			 */
			activity_write(BFHERE "PTY fork %d", errno);
			bf_notty();
			return;
		case 0:
			/* CHILD PROCESS */
			BFTRACE("PTY detach");
			setsid();
			return;
	}

	/* PARENT PROCESS */
	if (waitpid(pid, &status, 0) <= 0) {
		status = errno;
		activity_write(BFHERE "PTY waitpid %d", status);
	} else if (WIFEXITED(status)) {
		status = WEXITSTATUS(status);
		activity_write(BFHERE "PTY exit %d", status);
	} else if (WIFSIGNALED(status)) {
		status = WTERMSIG(status);
		activity_write(BFHERE "PTY kill -%d", status);
		kill(getpid(), status);
		sleep(2);
	} else {
		activity_write(BFHERE "PTY ??? %d", status);
	}
	exit(status);
}


static void attach_to_tty(int fd) {
#ifdef TIOCSCTTY
	if (ioctl(fd, TIOCSCTTY, NULL) < 0) {
		activity_write(BFHERE "PTY ioctl(TIOCSCTTY) %d", errno);
	} else {
		activity_write(BFHERE "PTY ioctl(TIOCSCTTY) ok");
	}
#endif /* TIOCSCTTY */
}


#ifdef HAVE_OPENPTY
int bf_openpty(void) {
	bf_command_state_t *state = &command_state;
	int ret;

	if (use_pipe())
		return create_pipe();

	detach_from_tty();
	ret = openpty(&state->mfd, &state->sfd, state->ptyname, NULL, NULL);
	if (ret < 0) { return ret; }

	convert_native_to_utf8(state->ptyname, CONVERT_ALL, state->buf, BFBUFSIZ);
	strcpy(state->ptyname, state->buf);
	attach_to_tty(state->sfd);
	return ret;
}


#else /* !HAVE_OPENPTY */

#ifdef ZOS
/*
 * Unfortunately, finding a pseudo-terminal on z/OS is painful.
 * It uses the old BSD-style pseudo-terminals, where you use
 * brute force to find an available master.
 */
int bf_openpty(void) {
	int *mfd = &command_state.mfd;
	int *sfd = &command_state.sfd;
	char *ptyname = command_state.ptyname;
	int i;

	if (use_pipe())
		return create_pipe();

	detach_from_tty();
	for (i=0; i<=9999; i++) {
		/* Look for a master that is available */
		sprintf(ptyname, "/dev/ptyp%04d", i);
		if ((*mfd = open(ptyname, O_RDWR)) < 0) {
			if (errno == ENOENT)
				return -1;
			continue;
		}

		/* See if the matching slave is available */
		ptyname[5] = 't';
		chown(ptyname, getuid(), -1);
		/* It should be appropriate to change this to mode 620, but
		 * the OS is not restoring it to 666 even though it is being
		 * smart enough to restore it to system ownership.
		 */ 
		/* chmod(ptyname, S_IRUSR | S_IWUSR | S_IWGRP); */
		if ((*sfd = open(ptyname, O_RDWR)) < 0) {
			/* Oops, got the master but not the slave */
			close(*mfd);
			continue;
		}

#ifdef HAVE_STROPTS_H
		if (ioctl(*sfd, I_LIST, NULL) > 0) {	/* streams support ? */
			(void)ioctl(*sfd, I_PUSH, "ptem");
			(void)ioctl(*sfd, I_PUSH, "ldterm");
		}
#endif /* HAVE_STROPTS_H */
		return 0;
	}

	/* We exhausted the entire pseudo-terminal namespace?! */
	errno = ENOENT;
	return -1;
}

#else /* !ZOS */

int bf_openpty(void) {
	bf_command_state_t *state = &command_state;
	int *mfd = &command_state.mfd;
	int *sfd = &command_state.sfd;
	char *ptyname = command_state.ptyname;
	int err, cloned = 0;
	char *sname;
#ifdef TIOCTRAP
	int one = 1;
#endif

	if (use_pipe())
		return create_pipe();

	if ((*mfd = open("/dev/ptmx", O_RDWR)) < 0 &&
			(*mfd = open("/dev/ptc", O_RDWR)) < 0)
	{
		if ((*mfd = open("/dev/ptym/clone", O_RDWR)) < 0)
			return -1;

		cloned = 1;
#ifdef TIOCTRAP
		ioctl(*mfd, TIOCTRAP, &one);
#endif
	}

	if ((sname = ptsname(*mfd)) != NULL) {
		strcpy(ptyname, sname);
		if (cloned || (!grantpt(*mfd) && !unlockpt(*mfd))) {
			if ((*sfd = open(ptyname, O_RDWR)) >= 0) {
#ifdef HAVE_STROPTS_H
				/* streams support? */
				if (!cloned && ioctl(*sfd, I_LIST, NULL) > 0) {
					if (ioctl(*sfd, I_FIND, "ldterm") != 1 &&
						ioctl(*sfd, I_FIND, "ldtty") != 1)
					{
						ioctl(*sfd, I_PUSH, "ptem");
						ioctl(*sfd, I_PUSH, "ldterm");
					}
				}
#endif /* HAVE_STROPTS_H */
				attach_to_tty(state->sfd);
				return 0;
			}
		}
	}

	err = errno;
	close(*mfd);
	*mfd = -1;
	errno = err;
	return -1;
}

#endif /* ZOS */
#endif /* HAVE_OPENPTY */


#ifdef DEBUG_PTY

int main(int argc, char *argv[]) {
	bf_command_state_t *state = &command_state;
	char **execArgv;
	int i, ret;

	if (argc < 3) {
		printf("Usage: %s command argv0 [argv1 [...]]\n", argv[0]);
		exit(1);
	}

	execArgv = calloc(argc, sizeof(char *));
	for (i=2; i<argc; i++)
		execArgv[i-2] = argv[i];
	execArgv[argc-2] = NULL;

	memset(state, 0, sizeof(*state));
	state->mfd = state->sfd = -1;
	strcpy(state->ptyname, "(unchanged)");

	ret = bf_openpty();
	if (ret != 0) {
		printf("bf_openpty failed: mfd=%d sfd=%d ptyname=%s ret=%d errno=%d\n",
			state->mfd, state->sfd, state->ptyname, ret, errno);
		exit(2);
	}

	execv(argv[1], execArgv);
	printf("exec failed: %s: errno=%d\n", argv[1], errno);
	exit(3);
}

#endif /* DEBUG_PTY */
